/*
 * Header functions for MockingJay II converters
 * SC5317A SC5318A
 *
 *	
 *	Copyright (c) 2022 SignalCore Inc.
 *	
 *	Rev 2.0.0
*/

#ifndef MOCKINGJAY2_FUNCTIONS_H__
#define MOCKINGJAY2_FUNCTIONS_H__

#ifdef __cplusplus
extern "C"
{
#endif

/*! Communication functions */
/*	Function to find the serial numbers of all SignalCore device with the same product ID
return:		The number of product devices found
input:		comm_interface enum {USB, RS232, PXI}
output:		2-D array (or pointers) to pass out the list serial numbers for devices found
Example, calling function could declare:
	char **serial_number_list;
	sNList = (char**)malloc(sizeof(char*)*50); // 50 serial numbers
	for (i=0;i<50; i++)
		searchNumberList[i] = (char*)malloc(sizeof(char)*(SCI_SN_LENGTH + 1)); 
and pass searchNumberList into the function.
*/
SCISTATUS mj2_SearchDevices(sci_comm_interface_t comm_interface,
	char** serial_number_list, int* number_devices);

/*	Same as above for LabVIEW calls
*/
SCISTATUS mj2_SearchDevicesLV(sci_comm_interface_t comm_interface,
	char* serial_number_list, int* number_devices);

/*	Function opens the target device.
	return:		pointer to device handle
	input: 		-comm_interface enum {USB, RS232, PXI}
				-rate ignored for PXI and USB. For RS232 0=>baud=57600, 1=>baud=115200
				-devSerialNum is the product serial number. Product number is available on
				the product label.
*/
SCISTATUS mj2_OpenDevice(sci_comm_interface_t comm_interface,
	char* devSerialNum, uint8_t rate, PHANDLE dev_handle);

/*	Function  closes the device associated with the handle.
	return:		error code
	input: 		device handle
*/
SCISTATUS mj2_CloseDevice(HANDLE dev_handle);


/* 	Register level access function prototypes 
	=========================================================================================
*/

/* 	Writing the register with via the USB device handle allocated by OpenDevice
	return: error code
	input: commandByte contains the target register address, eg 0x10 is the frequency register
	input: instructWord contains necessary data for the specified register address
*/
SCISTATUS mj2_RegWrite(HANDLE dev_handle,
							uint8_t command_byte, 
							uint64_t instruct_word);

/* 	Reading the register with via the USB device handle allocated by OpenDevice
	input: commandByte contains the target register address, eg 0x10 is the frequency register
	input: instructWord contains necessary data for the specified register address
	output: receivedWord is the return data request through the commandByte and instructWord
*/							
SCISTATUS mj2_RegRead(HANDLE dev_handle,
							uint8_t command_byte,
							uint64_t instruct_word,
							uint64_t *received_word);
							
/* 	Product configuration wrapper function prototypes 
	=========================================================================================
*/
/*	Initializes the device
	return: error code
	input: 		Mode	0: 	The device initializes to the power up state
						1:	The device reprograms all internal components to the current device 
							state
*/							
SCISTATUS mj2_InitDevice(HANDLE dev_handle, uint8_t mode);

/*	Sets the device RF port frequency
return: error code
input:	frequency in Hz. 
*/
SCISTATUS mj2_SetFrequency(HANDLE dev_handle, double frequency);

/*	Sets the device RF port frequency
return: error code
input:	frequency in Hz. 
*/
SCISTATUS mj2_SetIfFrequency(HANDLE dev_handle, double freq);

/*	Sets the device RF port frequency
return: error code
input:	frequency in Hz.
*/
SCISTATUS mj2_SetLoFrequency(HANDLE dev_handle, double freq);

/*	Set the synthesizer mode to determine the tuning behavior of the device
	return:	error code
	input:	pllLoopGain		0:	set to low loop bandwidth
							1:	set to normal loop bandwidth
							2:	set to high loop bandwidth
*/
SCISTATUS mj2_SetSynthMode(HANDLE dev_handle, uint8_t pllLoopGain);

/*	Set the LO parameters
	return:	error code
	input:	external_lo selects external LO, 0 = internal
			lo_mode 0=uses the RF, IF and spectral inversion to set LO, 1= set LO directly
			lo_doubler = manually set doubler on, only if externalLo = 1;
			ext_lo_path = selects either the front(0) or rear(1) LO ports, only if externalLo = 1
*/
SCISTATUS mj2_SetLoSource(HANDLE dev_handle, uint8_t external_lo,
	uint8_t lo_mode, uint8_t lo_doubler, uint8_t ext_lo_path);

/**	Sets the RF Preamplifier
return: error code
input:	enable
*/
SCISTATUS mj2_SetPreamp(HANDLE dev_handle, uint8_t enable);

/**	Sets the RF or IF attenuators
return: error code
input:	attenuator 0 RF; 1 IF
		atten_value dBm
*/
SCISTATUS mj2_SetAttenuator(HANDLE dev_handle,
								uint8_t attenuator,
								float atten_value);

/*	Set the signal path route of the converter
return:	error code
input	path parameters
*/
SCISTATUS mj2_SetSignalPath(HANDLE dev_handle,
	const signal_path_params_t* signal_path_params);

/* Set the device to automatically set the device gain using the gain_params*/
/* If autoConvEnable is set enable (1), the device will automatically set up the
* device to achieve the required conversion gain, name the difference between
* IF out - RF in.
*/
SCISTATUS mj2_SetAutoConversion(HANDLE deviceHandle,
	const gain_params_t *gain_params,	uint8_t auto_conv_enable);

/*	Puts the device in power standby mode.
	return: error code
	input: standbyStatus	0:	Take deviceout off power standby
							1:	The put deviceinto power standby
*/
SCISTATUS mj2_SetStandby(HANDLE dev_handle, uint8_t standby_status);

/*	Set the reference clock configurations
	return:	error code
	input:	lock_ext_enable	enables the device to lock to an external source. If the source
							not available error code returns. The device will not attempt
							to lock and waits for source.
	input:  pxi10_clk_enable  enable PXI10Clk output for PXIe devices
*/
SCISTATUS mj2_SetReferenceClock(HANDLE dev_handle, uint8_t pxi10_clk_enable, uint8_t lock_ext_enable);

/*	Sets the value of the Clock reference DAC and hence adjusts reference clock frequency
	return: error code
	input: 	dac_value	a 14 bit word
*/
SCISTATUS mj2_SetReferenceDac(HANDLE dev_handle, uint32_t dac_value);

/**	Store the current state of the signal source into EEPROM as the default startup state
return: error code
input: 	none
*/
SCISTATUS mj2_SetAsDefault(HANDLE dev_handle);

/*	Write single byte to the user EEPROM address
return: error code
input: 	mem_add		the address of the EEPROM memory to write to
input:	byte_data	the byte data to write to the specified memory address
*/
SCISTATUS mj2_WriteUserEeprom(HANDLE dev_handle,
	uint32_t mem_add,
	uint8_t byte_data);

/**	Self Cal of the synthesizer to align the VCOs
return: error code
*/
SCISTATUS mj2_SetSynthSelfCal(HANDLE dev_handle);

/* Set (Store) the autogain parameter used to compute the attenuator setting*/
/* This function does nothing to the device accept to store the parameters use
	by the computational functions into device onboard memory. The user could 
	retrive these parameters for later use.
*/
SCISTATUS mj2_SetGainCalcParams(HANDLE deviceHandle, gain_params_t *gain_params);

/* Product Export Query (Read) function prototypes */
/*----------------------------------------------------------------------------------------------- */

/*	Function retrives the rf parameters such as rf, if, lo frequencies and other rf components
return:	error code
output:	rf_parameters
*/
SCISTATUS mj2_GetRfParameters(HANDLE dev_handle, rf_params_t *rf_params);

/*	Function retrives current temperature of the device
return:	error code
output:	temperature
*/
SCISTATUS mj2_GetTemperature(HANDLE dev_handle, float *temp);

/*	Function retrives the device status - PLL locks status, ref clk config, etc see device_status_t type
return:	error code
output:	device status
*/
SCISTATUS mj2_GetDeviceStatus(HANDLE dev_handle, device_status_t *device_status);

/*	Function fetches the device info
return:	error code
output:	deviceInfo		device information structure
*/
SCISTATUS mj2_GetDeviceInfo(HANDLE dev_handle, device_info_t *device_info);

/* Function fetches the device calibration data stored in memory
return:	error code
output:	pointer to calibration data
*/
SCISTATUS mj2_GetCalData(HANDLE dev_handle, cal_data_t **cal_data);

/* Function calculates the conversion gain of the device with its current setup
return:	error code
output:	conversion gain
*/
SCISTATUS mj2_CalcGain(HANDLE dev_handle, float* conversion_gain);

/* Function the maximun of the device for the RF frequency and IF frequency choosen
return:	error code
output:	provides the max possible gain without the preamp on and pre_amp_gain values \
		at the current frequencies.
*/
SCISTATUS mj2_CalcMaxGain(HANDLE dev_handle, float* max_gain, float* pre_amp_gain);

/* Function calculates the best attenuation based on the entered auto_gain_params
return:	error code
output:	attenuator values
*/ 
SCISTATUS mj2_CalcAttenValues(HANDLE dev_handle, attenuator_t* atten);

#ifdef __cplusplus
}
#endif

#	endif  /* MOCKINGJAY2_FUNCTIONS_H__ */	