/*
 *
  ******************************************************************************
 * @attention
 *
 * Copyright (c) 2020-2023 SignalCore.
 * All rights reserved.
 *
 * This software is licensed under terms that can be found in the LICENSE file
 * in the root directory of this software component.
 * If no LICENSE file comes with this software, it is provided AS-IS.
 *
 ******************************************************************************
 */

/*
 * Advance Mockingjay II functions
 *	
 *	Rev 2.0
*/

#include <stdlib.h>
#include <malloc.h>
#include <stdint.h>
#include <string.h>
#include <math.h>

#include "sci_types.h"
#include "sci_errors.h"
#include "sci_device.h"
#include "sci_math.h"

#include "mj2_comm.h"
#include "mj2_defs.h"
#include "mj2_regs.h"
#include "mj2_functions.h"
#include "mj2_advanceregs.h"
#include "mj2_device.h"
#include "mj2_advance.h" 
#include <math.h>

SCISTATUS mj2adv_SaveAutoConvParams(HANDLE dev_handle,
	const gain_params_t* gain_params, uint8_t auto_conv_enable)
{
	uint64_t tmpU64 = 0x00;
	uint8_t tmpU8;
	 
	tmpU64 = ((uint64_t)auto_conv_enable & 0x01)
		| (((uint64_t)gain_params->auto_amp_ctrl & 0x01) << 1)
		| (((uint64_t)gain_params->linearity_mode & 0x07) << 2);


	tmpU8 = (uint8_t)floor(fabsf(gain_params->rf_level));
	if (gain_params->rf_level < 0)
		tmpU8 |= 0x80;
	else
		tmpU8 |= 0x00;
	tmpU64 |= ((uint64_t)tmpU8 << 24);

	tmpU8 = (uint8_t)floor(fabsf(gain_params->mixer_level));
	if (gain_params->mixer_level < 0)
		tmpU8 |= 0x80;
	else
		tmpU8 |= 0x00;
	tmpU64 |= ((uint64_t)tmpU8 << 16);

	tmpU8 = (uint8_t)round(fabsf(gain_params->if_level));
	if (gain_params->if_level < 0)
		tmpU8 |= 0x80;
	else
		tmpU8 |= 0x00;
	tmpU64 |= ((uint64_t)tmpU8 << 8);

	return mj2_RegWrite(dev_handle, AUTO_CALC_GAIN, tmpU64);
}

SCISTATUS mj2adv_AutoCalcAttenValues(HANDLE dev_handle, attenuator_t* atten)
{
	SCISTATUS status;
	double rfAtten, ifAtten;
	double if3Atten1Default;
	double pathGain, rfAbsGain, ampGain;
	double tempGain, ifGain, deltaTemp, tempcoeff1, tempcoeff2;
	int rfIndex, ifIndex;
	double mixer_level = -20;
	PSCIDEVICE sci_device = (PSCIDEVICE)dev_handle;
	device_data_t* device_data = (device_data_t*)sci_device->device_data;
	cal_data_t* cal_data;

	if (!device_data->cal_data_read)
	{
		status = mj2adv_FetchCalData(dev_handle);  // if cal data is not read yet, allocate and fetch
		if (status != EXIT_SUCCESS) 
			return status;
	}

	cal_data = (cal_data_t*)device_data->cal_data; //useful only after allocation

	if (!device_data->gain_params_set)
	{
		status = mj2adv_FetchGainCalcParams(dev_handle);  // if cal data is not read yet, allocate and fetch
		if (status != EXIT_SUCCESS)
			return status;
	}

	switch (device_data->gain_params.linearity_mode)
	{
	case 0:
	default:
		if (device_data->gain_params.mixer_level > 0.00)
			mixer_level = 0;
		else
			mixer_level = device_data->gain_params.mixer_level;
		if3Atten1Default = 4.00;
		break;
	case 1:
		mixer_level = -20.00;
		if3Atten1Default = 4.00;
		break;
	case 2:
		mixer_level = -10.00;
		if3Atten1Default = 0.00;
		break;
	case 3:
		mixer_level = 0.0;
		if3Atten1Default = 0.00;
		break;
	case 4:
		mixer_level = -25.00;
		if3Atten1Default = 4.00;
		break;
	case 5:
		mixer_level= -30.00;
		if3Atten1Default = 4.00;
		break;
	}
	
	// work out the rf absolute gain
	rfIndex = (int)floor((device_data->rf_params.rf_frequency - 6e9) / 250e6);
	ifIndex = (int)floor(device_data->rf_params.if_frequency / 100e6);
	if (rfIndex < 0) rfIndex = 0;
	if (ifIndex < 0) ifIndex = 0;


	if (device_data->rf_params.path_params.invert_spectrum)
		rfAbsGain = cal_data->rf_lsb_abs_gain_cal[rfIndex];
	else
		rfAbsGain = cal_data->rf_usb_abs_gain_cal[rfIndex];
	 
	ampGain = cal_data->rf_amp_gain_cal[rfIndex];

	if (device_data->rf_params.rf_frequency > 20e9)
	{
		tempcoeff1 = cal_data->temp_coeff[4];
		tempcoeff2 = cal_data->temp_coeff[5];
	}
	else if (device_data->rf_params.rf_frequency > 13e9)
	{
		tempcoeff1 = cal_data->temp_coeff[2];
		tempcoeff2 = cal_data->temp_coeff[3];
	}
	else
	{
		tempcoeff1 = cal_data->temp_coeff[0];
		tempcoeff2 = cal_data->temp_coeff[1];
	}

	deltaTemp = (double)device_data->temperature - (double)cal_data->cal_temp;

	rfAtten = (double)(device_data->gain_params.rf_level - mixer_level);

	//determine what to do with the preamp
	if (device_data->gain_params.auto_amp_ctrl)
	{
		if ((rfAtten < -10) && (device_data->gain_params.linearity_mode == 3))
		{
			device_data->rf_params.path_params.rf_amp_enable = 1;
			rfAtten += ampGain;
		}
		else if (rfAtten < -15.0)
		{
			device_data->rf_params.path_params.rf_amp_enable = 1;
			rfAtten += ampGain;
		}
		else
		{
			device_data->rf_params.path_params.rf_amp_enable = 0;
		}
	}
	else
	{
		if (device_data->rf_params.path_params.rf_amp_enable == 1)
			rfAtten += ampGain;
	}

	if (rfAtten < 0) rfAtten = 0;  //attenuation can only be positive
	// Determine the Absolute Gain of the choosen path 

	pathGain = (double)device_data->gain_params.if_level - (double)device_data->gain_params.rf_level;

	if (device_data->rf_params.path_params.rf_amp_enable)
		rfAbsGain += ampGain;  // total RF gain

	// Calc the IF Atten assuming IF is configure as cal reference
	ifAtten = rfAbsGain - pathGain - rfAtten;

	// If IF is not configure as the Cal reference state, need to get corrections
	ifGain = cal_data->if_rel_gain_cal[ifIndex];

	tempGain = tempcoeff1 * deltaTemp + tempcoeff2 * pow(deltaTemp, 2);

	// adjust the attenuation with the gains
	ifAtten = ifAtten + tempGain + ifGain;

	if (ifAtten < 0) ifAtten = 0;
	ifAtten = (ifAtten + rfAtten) - round(rfAtten);
	rfAtten = round(rfAtten);
	ifAtten = round(ifAtten);

	if (rfAtten > CALATTENSTEPS) rfAtten = CALATTENSTEPS;
	if (ifAtten > CALATTENSTEPS) ifAtten = CALATTENSTEPS;

	atten->rf_atten_value = rfAtten;
	atten->if_atten_value = ifAtten;

	return EXIT_SUCCESS;
}

SCISTATUS mj2adv_ReadEeprom(HANDLE dev_handle,
	uint8_t eeprom,
	uint32_t memStartAdd,
	uint32_t len,
	uint8_t* byteData)
{
	int32_t readCycles, count, i;
	SCISTATUS status;
	uint64_t receivedWord = 0;
	uint8_t* dataBuf;

	if (len < 1)
	{
		return SCI_ERROR_INVALID_ARGUMENT;
	}
	
	if (byteData == NULL)
		return SCI_ERROR_MEM_UNALLOCATE;
	else
	{
		dataBuf = (uint8_t*)malloc(sizeof(uint8_t) * len + 8);
		if (dataBuf == NULL)
			return SCI_ERROR_MEM_UNALLOCATE;

		readCycles = (int32_t)(floor((float)len / 8.0) + 1.0);

		count = 0;
		while (count < readCycles)
		{
			status = mj2_RegRead(dev_handle, CAL_EEPROM_READ, ((uint64_t)eeprom << 16) | (uint64_t)(memStartAdd + (count * 8) & 0xFFFF), &receivedWord);
			if (status != SCI_SUCCESS) return status;

			for (i = 0; i < 8; i++)
			{
				dataBuf[i + count * 8] = (uint8_t)((receivedWord >> (i * 8)) & 0xFF);
				if ((i + count * 8) == (len - 1)) break;
			}
			count++;
		}

		for (i = 0; i < len; i++)
			byteData[i] = dataBuf[i];
	}

	free(dataBuf);
	dataBuf = NULL;
	return SCI_SUCCESS;
}

SCISTATUS mj2adv_FetchRawCalData(HANDLE dev_handle, uint8_t *raw_cal_data)
{
	return mj2adv_ReadEeprom(dev_handle, CALEEPROM,
		CALDATASTARTADD, CALDATALEN, raw_cal_data);
}

SCISTATUS mj2adv_AllocateCalDataMemory(cal_data_t *cal_data)
{
	//* 	AllocateCalData allocates memory for cakibration data
	//

	int i;
	if (cal_data)
	{
		cal_data->temp_coeff = (float*)malloc(TEMPCOLEN * sizeof(float));  // allocate memory of temp coefficients
		cal_data->if_cal_freq = (float*)malloc(IFCALFREQLEN * sizeof(float));
		cal_data->if_rel_gain_cal = (float*)malloc(IFCALFREQLEN * sizeof(float));
		cal_data->if_atten_cal = (float*)malloc(CALATTENSTEPS * sizeof(float*));
		cal_data->bp_rf_cal_freq = (float*)malloc(BYPASSFREQLEN * sizeof(float));
		cal_data->bp_abs_gain_cal = (float*)malloc(BYPASSFREQLEN * sizeof(float));
		cal_data->rf_cal_freq = (float*)malloc(RFCALFREQLEN * sizeof(float));
		cal_data->rf_lsb_abs_gain_cal = (float*)malloc(RFCALFREQLEN * sizeof(float));
		cal_data->rf_usb_abs_gain_cal = (float*)malloc(RFCALFREQLEN * sizeof(float));
		cal_data->rf_amp_gain_cal = (float*)malloc(RFCALFREQLEN * sizeof(float));
		cal_data->rf_atten_cal = (float**)malloc(CALATTENSTEPS * sizeof(float*));

		if(cal_data->rf_atten_cal)
		for (i = 0; i < CALATTENSTEPS; i++)
			cal_data->rf_atten_cal[i] = (float*)malloc(RFCALFREQLEN * sizeof(float));

	}

	return EXIT_SUCCESS;
}

SCISTATUS mj2adv_DeallocateCalDataMemory(cal_data_t* cal_data)
{
	//Deallocates memory allocated 
	int i;

	free(cal_data->temp_coeff);
	cal_data->temp_coeff = NULL;

	free(cal_data->if_cal_freq);
	cal_data->if_cal_freq = NULL;

	free(cal_data->if_rel_gain_cal);
	cal_data->if_rel_gain_cal = NULL;

	free(cal_data->if_atten_cal);
	cal_data->if_atten_cal = NULL;

	free(cal_data->bp_rf_cal_freq);
	cal_data->bp_rf_cal_freq = NULL;

	free(cal_data->bp_abs_gain_cal);
	cal_data->bp_abs_gain_cal = NULL;

	free(cal_data->rf_cal_freq);
	cal_data->rf_cal_freq = NULL;

	free(cal_data->rf_lsb_abs_gain_cal);
	cal_data->rf_lsb_abs_gain_cal = NULL;

	free(cal_data->rf_usb_abs_gain_cal);
	cal_data->rf_usb_abs_gain_cal = NULL;

	free(cal_data->rf_amp_gain_cal);
	cal_data->rf_amp_gain_cal = NULL;

	for (i = 0; i < CALATTENSTEPS; i++)
		free(cal_data->rf_atten_cal[i]);
	free(cal_data->rf_atten_cal);
	cal_data->rf_atten_cal = NULL;

	free(cal_data);
	cal_data = NULL;

	return EXIT_SUCCESS;
}

SCISTATUS mj2adv_RawToCalData(const uint8_t *raw_cal_data,
	cal_data_t *cal_data)
{
	int i = 0, j = 0;
	//	float tmp_f32;
	if ((raw_cal_data == NULL) || (cal_data == NULL))
		return SCI_ERROR_MEM_UNALLOCATE;

	// Get Temperature related
	// Calibration temperature 
	cal_data->cal_temp = *(float*)&raw_cal_data[CALTEMP_ADD - CALDATASTARTADD];
	// Temp cal frequencies 
	//Temperature Coefficients 1D arrary [2]
	for (i = 0; i < TEMPCOLEN; i++)
	{
		cal_data->temp_coeff[i] =
			*(float*)&raw_cal_data[(TEMPCOEFF_ADD - CALDATASTARTADD) + i * 4];
	}

	for (i = 0; i < IFCALFREQLEN; i++)
	{
		cal_data->if_cal_freq[i] =
			*(float*)&raw_cal_data[(IFCALFREQ_ADD - CALDATASTARTADD) + 4 * i];
	//* IF3 relative response for Higher LPF *
		cal_data->if_rel_gain_cal[i] =
			*(float*)&raw_cal_data[(IFRELGAIN_ADD - CALDATASTARTADD) + 4 * i];
	}

	for (i = 0; i < CALATTENSTEPS; i++)
	{
		cal_data->if_atten_cal[i] = *(float*)&raw_cal_data[(IFATTENCAL_ADD - CALDATASTARTADD) + i * 4];
	}

	//*** Bypass RF conversion ***
	//* Bypass conversion cal frequencies *								
	for (i = 0; i < BYPASSFREQLEN; i++)
	{
		cal_data->bp_rf_cal_freq[i] =
			*(float*)&raw_cal_data[(BPCALFREQ_ADD - CALDATASTARTADD) + 4 * i];
	//* Bypass conversion absolute gain*
		cal_data->bp_abs_gain_cal[i] =
			*(float*)&raw_cal_data[(BPABSGAIN_ADD - CALDATASTARTADD) + 4 * i];
	}

	//*** RF Calibration ***
	//* RF cal frequencies *
	for (i = 0; i < RFCALFREQLEN; i++)
	{
		cal_data->rf_cal_freq[i] =
			*(float*)&raw_cal_data[(RFCALFREQ_ADD - CALDATASTARTADD) + 4 * i];

	//* RF non-inverting absolute gain at each frequency *
		cal_data->rf_usb_abs_gain_cal[i] =
			*(float*)&raw_cal_data[(RFNONINVGAIN_ADD - CALDATASTARTADD) + 4 * i];
	
		cal_data->rf_lsb_abs_gain_cal[i] =
			*(float*)&raw_cal_data[(RFINVGAIN_ADD - CALDATASTARTADD) + 4 * i];
	
	//* RF relative amplifier gain at each frequency *
		cal_data->rf_amp_gain_cal[i] =
			*(float*)&raw_cal_data[(RFAMPGAIN_ADD - CALDATASTARTADD) + 4 * i];
	}

	//* RF relative Atten 1 cal *
	for (i = 0; i < CALATTENSTEPS; i++)
	{
		for (j = 0; j < RFCALFREQLEN; j++)
		{
			cal_data->rf_atten_cal[i][j] =
				*(float*)&raw_cal_data[(RFATTENCAL_ADD - CALDATASTARTADD) + j * 4 + i * RFCALFREQLEN * 4];
		}
	}

	return SCI_SUCCESS;
}

SCISTATUS mj2adv_FetchCalData(HANDLE dev_handle)
{
	SCISTATUS status = 0;
	PSCIDEVICE sci_device = (PSCIDEVICE)dev_handle;
	device_data_t* device_data = (device_data_t*)sci_device->device_data; //cast the void* device_data to a calData_t* struct

	uint8_t* rawData = (uint8_t*)malloc((CALDATALEN + 8) * sizeof(char));
	cal_data_t* cal_data = (cal_data_t*)malloc(sizeof(cal_data_t));

	status = mj2adv_FetchRawCalData(dev_handle,  rawData);
	if (status != SCI_SUCCESS) goto end;

	status = mj2adv_AllocateCalDataMemory(cal_data);
	if (status != SCI_SUCCESS) goto clearAllocate;

	status = mj2adv_RawToCalData(rawData, cal_data);
	if (status != SCI_SUCCESS) goto clearAllocate;

	device_data->cal_data = cal_data;
	device_data->cal_data_read = 1;
	goto end;
	
clearAllocate:
	mj2adv_DeallocateCalDataMemory(cal_data);
	free(cal_data);
end:
	free(rawData);
	rawData = NULL;

	return status;
}

SCISTATUS mj2adv_ReleaseCalData(HANDLE dev_handle)
{
	SCISTATUS status = 0;
	PSCIDEVICE sci_device = (PSCIDEVICE)dev_handle; //cast the void* dev_handle memory back to a SCIDEVICE* struct
	device_data_t* device_data = (device_data_t*)sci_device->device_data; //cast the void* device_data to a device_data_t* struct

	if (device_data->cal_data != NULL) //if memory not empty
	{
		status = mj2adv_DeallocateCalDataMemory(device_data->cal_data);
		device_data->cal_data = NULL;
	}
	device_data->cal_data_read = 0;

	return status;
}

SCISTATUS mj2adv_FetchRfParameters(HANDLE dev_handle)
{

	SCISTATUS status;
	uint64_t received_word = 0;

	PSCIDEVICE sci_device = (PSCIDEVICE)dev_handle; 
	device_data_t* device_data = (device_data_t*)sci_device->device_data; 

	status = mj2_RegRead(dev_handle, GET_DEVICE_PARAM, 0x00, &received_word);
	if (status != EXIT_SUCCESS) return status;
	device_data->rf_params.rf_frequency = (double)received_word * 0.001;

	status = mj2_RegRead(dev_handle, GET_DEVICE_PARAM, 0x01, &received_word);
	if (status != EXIT_SUCCESS) return status;
	device_data->rf_params.if_frequency = (double)received_word * 0.001;

	status = mj2_RegRead(dev_handle, GET_DEVICE_PARAM, 0x02, &received_word);
	if (status != EXIT_SUCCESS) return status;
	device_data->rf_params.lo_frequency = (double)received_word * 0.001;

	status = mj2_RegRead(dev_handle, GET_DEVICE_PARAM, 0x03, &received_word);
	if (status != EXIT_SUCCESS) return status;
	device_data->rf_params.atten.if_atten_value = (float)(received_word & 0xFF) * 0.25;
	device_data->rf_params.atten.rf_atten_value = (float)((received_word >> 8) & 0xFF) * 0.25;
	device_data->rf_params.path_params.bypass_conversion = (uint8_t)((received_word >> 16) & 0x01);
	device_data->rf_params.path_params.rf_amp_enable = (uint8_t)((received_word >> 17) & 0x01);
	device_data->rf_params.path_params.if_out_enable = (uint8_t)((received_word >> 18) & 0x01);
	device_data->rf_params.path_params.invert_spectrum = (uint8_t)((received_word >> 19) & 0x01);
	
	device_data->rf_params_read = 1;

	return EXIT_SUCCESS;
}

SCISTATUS mj2adv_FetchGainCalcParams(HANDLE dev_handle)
{
	int status;
	uint8_t tmp_u8;
	uint64_t received_word = 0;

	PSCIDEVICE sci_device = (PSCIDEVICE)dev_handle;
	device_data_t* device_data = (device_data_t*)sci_device->device_data;

	status = mj2_RegRead(dev_handle, GET_DEVICE_PARAM, 8, &received_word);
	if (status != SCI_SUCCESS) return status;

	tmp_u8 = (uint8_t)((received_word >> 24) & 0xFF);
	if (tmp_u8 & 0x80) device_data->gain_params.rf_level = -(float)(tmp_u8 & 0x7F);
	else device_data->gain_params.rf_level = (float)(tmp_u8 & 0x7F);

	tmp_u8 = (uint8_t)((received_word >> 16) & 0xFF);
	if (tmp_u8 & 0x80) device_data->gain_params.mixer_level = -(float)(tmp_u8 & 0x7F);
	else device_data->gain_params.mixer_level = (float)(tmp_u8 & 0x7F);

	tmp_u8 = (uint8_t)((received_word >> 8) & 0xFF);
	if (tmp_u8 & 0x80) device_data->gain_params.if_level = -(float)(tmp_u8 & 0x7F);
	else device_data->gain_params.if_level = (float)(tmp_u8 & 0x7F);

	device_data->gain_params.linearity_mode = (uint8_t)(received_word >> 1) & 0x07;
	device_data->gain_params.auto_amp_ctrl = (uint8_t)(received_word & 0x01);

	device_data->gain_params_set = 1;

	return status;
}

SCISTATUS mj2adv_FetchStartUpParams(HANDLE dev_handle)
{
	if (dev_handle == NULL)
		return SCI_ERROR_INVALID_DEVICE_HANDLE;
	SCISTATUS status;
	uint64_t received_word = 0;
	uint32_t tmp_u32 = 0;

	PSCIDEVICE sci_device = (PSCIDEVICE)dev_handle;
	device_data_t* device_data = (device_data_t*)sci_device->device_data;

	status = mj2_RegRead(dev_handle, GET_DEVICE_STATUS, 0, &received_word);
	if (status != EXIT_SUCCESS) return status;

	device_data->auto_conv_enable = (uint8_t)((received_word & 0x0000200000) >> 21);
	
	status = mj2_RegRead(dev_handle, GET_TEMPERATURE, 0, &received_word);
	if (status != EXIT_SUCCESS) return status;
	tmp_u32 = (uint32_t)received_word;
	device_data->temperature = *(float*)&tmp_u32; //copy temp into device_data memory

	return EXIT_SUCCESS;
}

SCISTATUS mj2adv_FetchDeviceInfo(HANDLE dev_handle)
{
	if (dev_handle == NULL)
		return SCI_ERROR_INVALID_DEVICE_HANDLE;
	SCISTATUS status;
	uint32_t tmpU32;
	uint64_t received_word = 0;

	PSCIDEVICE sci_device = (PSCIDEVICE)dev_handle;
	device_data_t* device_data = (device_data_t*)sci_device->device_data;
												   //Product Serial Number 
	status = mj2_RegRead(dev_handle, GET_DEVICE_INFO, DEVINFOSNINT, &received_word);
	device_data->info.product_sn = (uint32_t)received_word;
	tmpU32 = (uint8_t)(received_word >> 32);
	device_data->info.dev_interface = 0;
	if (tmpU32 & 0x01) device_data->info.dev_interface = 1;
	if ((tmpU32 & 0x02) && (tmpU32 & 0x04)) device_data->info.dev_interface = 2;
	if ((tmpU32 & 0x02) && (tmpU32 & 0x08)) device_data->info.dev_interface = 3;

	status = mj2_RegRead(dev_handle, GET_DEVICE_INFO, DEVINFOREV, &received_word);
	tmpU32 = (uint32_t)received_word;
	device_data->info.fw_rev = *(float*)&tmpU32;
	tmpU32 = (uint32_t)(received_word >> 32);
	device_data->info.hw_rev = *(float*)&tmpU32;

	status = mj2_RegRead(dev_handle, GET_DEVICE_INFO, DEVINFODATE, &received_word);

	device_data->info.man_date.year = (uint16_t)(received_word >> 48);
	device_data->info.man_date.month = (uint8_t)(received_word >> 40);
	device_data->info.man_date.day = (uint8_t)(received_word >> 32);
	device_data->info.cal_date.year = (uint16_t)(received_word >> 16);
	device_data->info.cal_date.month = (uint8_t)(received_word >> 8);
	device_data->info.cal_date.day = (uint8_t)(received_word);

	device_data->info_read = 1;

	return EXIT_SUCCESS;
}


