/*
 * Mocking Jay II communication functions
 *	
 * Copyright (c) SignalCore Inc.
 *	
 *	Rev 2.0
*/

#ifdef __cplusplus
extern "C"
{
#endif

#include <stdlib.h>
#include <malloc.h>
#include <stdint.h>
#include <string.h>

#include "sci_types.h"
#include "sci_errors.h"
#include "sci_device.h"
#include "sci_math.h"

#include "sci_pci_device.h"
#include "sci_usb_device.h"
#include "sci_rs232_device.h"

#include "mj2_comm.h"
#include "mj2_defs.h"
#include "mj2_regs.h"
#include "mj2_functions.h"
#include "mj2_advanceregs.h"
#include "mj2_device.h"
#include "mj2_advance.h"


// pointer functions to select the device interface

typedef SCISTATUS (*pSearchDevices)(char **devSerialNum, uint32_t* number_devices);
typedef SCISTATUS (*pSearchDevicesLV)(char *devSerialNum, uint32_t* number_devices);
typedef SCISTATUS (*pDeviceOpen)(char *devSerialNum, uint8_t rate, PHANDLE dev_handle);
typedef SCISTATUS (*pDeviceWrite)(HANDLE dev_handle, uint8_t *data, int len);
typedef SCISTATUS (*pDeviceRead)(HANDLE dev_handle, uint8_t *data, int len);
typedef SCISTATUS (*pDeviceClose)(HANDLE dev_handle);

/* Begin Implemenation of local functions */
/* *********************************************************************************** */
SCISTATUS WriteDevice(HANDLE dev_handle, uint8_t *data, int len)
{
	
	if (dev_handle == NULL)
		return SCI_ERROR_INVALID_DEVICE_HANDLE;

//	pDeviceWrite device_write_fp;
	PSCIDEVICE sciDevice = (PSCIDEVICE)dev_handle;

	if(sciDevice->device_type == USB_INT)
//		device_write_fp = &UsbWrite;
		return UsbWrite(sciDevice->handle, data, len);
	else if (sciDevice->device_type == RS232_INT)
//		device_write_fp = &Rs232Write;
		return Rs232Write(sciDevice->handle, data, len);
	else if (sciDevice->device_type == PCI_INT)
//		device_write_fp = &PciWrite;
		return PciWrite(sciDevice->handle, data, len);
	else
		return SCI_ERROR_INVALID_DEVICE_HANDLE;
	
//	return device_write_fp(sciDevice->handle, data, len);
}

SCISTATUS ReadDevice(HANDLE dev_handle, uint8_t *data, int len)
{
	if (dev_handle == NULL)
		return SCI_ERROR_INVALID_DEVICE_HANDLE;

//	pDeviceRead device_read_fp;
	PSCIDEVICE sciDevice = (PSCIDEVICE)dev_handle;

	if (sciDevice->device_type == USB_INT)
		//		device_read_fp = &UsbRead;
		return UsbRead(sciDevice->handle, data, len);
	else if (sciDevice->device_type == RS232_INT)
		//		device_read_fp = &Rs232Read;
		return Rs232Read(sciDevice->handle, data, len);
	else if (sciDevice->device_type == PCI_INT)
		//		device_read_fp = &PciRead;
		return PciRead(sciDevice->handle, data, len);
	else
		return SCI_ERROR_INVALID_DEVICE_HANDLE;


//	return device_read_fp(sciDevice->handle, data, len);
}

/* 	Begin Impementation of export functions 
****************************************************************************************************************
	
Begin USB device session related functions
=================================================================================================================
*/
SCISTATUS mj2_SearchDevices(sci_comm_interface_t comm_interface, 
	char **serial_number_list, int *number_devices)
{
	pSearchDevices device_search_fp;

	if (comm_interface == USB_INT)
		device_search_fp = &UsbSearchDevices;
	else if (comm_interface == RS232_INT)
		device_search_fp = &Rs232SearchDevices;
	else if (comm_interface == PCI_INT)
		device_search_fp = &PciSearchDevices;
	else
		return SCI_ERROR_INVALID_DEVICE;

	return device_search_fp(serial_number_list, number_devices);
}


SCISTATUS mj2_SearchDevicesLV(sci_comm_interface_t comm_interface, 
	char *serial_number_list, int *number_devices)
{
	pSearchDevicesLV device_search_lv_fp;

	if (comm_interface == USB_INT)
		device_search_lv_fp = &UsbSearchDevicesLV;
	else if (comm_interface == RS232_INT)
		device_search_lv_fp = &Rs232SearchDevicesLV;
	else if (comm_interface == PCI_INT)
		device_search_lv_fp = &PciSearchDevicesLV;
	else
		return SCI_ERROR_INVALID_DEVICE;

	return device_search_lv_fp(serial_number_list, number_devices);
}

SCISTATUS mj2_OpenDevice(sci_comm_interface_t comm_interface, char *devSerialNum, uint8_t rate, PHANDLE dev_handle)
{
	pDeviceOpen device_open_fp;
	pDeviceClose device_close_fp;

	SCISTATUS status;
	device_data_t* device_data = (device_data_t*)malloc(sizeof(device_data_t));
	PSCIDEVICE sciDevice = (PSCIDEVICE)malloc(sizeof(SCIDEVICE));

	if (comm_interface == USB_INT)
	{
		device_open_fp = &UsbOpenDevice;
		device_close_fp = &UsbCloseDevice;
	}
	else if (comm_interface == RS232_INT)
	{
		device_open_fp = &Rs232OpenDevice;
		device_close_fp = &Rs232CloseDevice;
	}
	else if (comm_interface == PCI_INT)
	{
		device_open_fp = &PciOpenDevice;
		device_close_fp = &PciCloseDevice;
	}
	else
	{
		free(device_data);
		free(sciDevice);
		return SCI_ERROR_INVALID_DEVICE;
	}

	if (sciDevice)
	{
		status = device_open_fp(devSerialNum, rate, &sciDevice->handle);
		if (status != EXIT_SUCCESS)
		{
			free(device_data);
			free(sciDevice);
			return status;
		}

		if (device_data)
		{
			device_data->cal_data = NULL;
			device_data->cal_data_read = 0;
			device_data->rf_params_read = 0;
			device_data->gain_params_set = 0;
			device_data->info_read = 0;
		}
		sciDevice->device_data = (void*)device_data;
		sciDevice->device_type = comm_interface;


		status = mj2adv_FetchRfParameters((HANDLE)sciDevice);
		if (status != SCI_SUCCESS)
		{
			device_close_fp(sciDevice->handle);
			*dev_handle = NULL;
			return status;
		}

		if (device_data)
		{
			device_data->last_rf = device_data->rf_params.rf_frequency;
			device_data->last_if = device_data->rf_params.if_frequency;
			device_data->last_amp_state = device_data->rf_params.path_params.rf_amp_enable;
		}

		status = mj2adv_FetchGainCalcParams((HANDLE)sciDevice);
		if (status != SCI_SUCCESS)
		{
			device_close_fp(sciDevice->handle);
			*dev_handle = NULL;
			return status;
		}

		status = mj2adv_FetchStartUpParams((HANDLE)sciDevice);
		if (status != SCI_SUCCESS)
		{
			device_close_fp(sciDevice->handle);
			*dev_handle = NULL;
			return status;
		}

		*dev_handle = (HANDLE)sciDevice;
		return EXIT_SUCCESS;
	}
	else
		return SCI_ERROR_DEVICE_UNALLOCATED;
}

SCISTATUS mj2_CloseDevice(HANDLE dev_handle)
{
	if (dev_handle == NULL)
		return SCI_ERROR_INVALID_DEVICE_HANDLE;
	SCISTATUS status;
	pDeviceClose close_device_fp;
	PSCIDEVICE sci_device = (PSCIDEVICE)dev_handle;
	device_data_t* device_data = (device_data_t*)sci_device->device_data;
	cal_data_t* cal_data = (cal_data_t*)device_data->cal_data;
	gain_params_t auto_gain_params = device_data->gain_params;

//	printf("gain_params linearity_mode is: %d", device_data->gain_params.linearity_mode);
//	printf("path_params if filter is: %d", device_data->path_params.if_filter_value);
	status = mj2adv_SaveAutoConvParams(dev_handle, &auto_gain_params, device_data->auto_conv_enable); //save the setting to the device
	if (sci_device->device_type == USB_INT)
		close_device_fp = &UsbCloseDevice;  //let the pointer function point to the USB close
	else if (sci_device->device_type == RS232_INT)
		close_device_fp = &Rs232CloseDevice; //let the pointer function point to the RS232 close
	else if (sci_device->device_type == PCI_INT)
		close_device_fp = &PciCloseDevice; //let the pointer function point to the PCI close
	else
		return SCI_ERROR_INVALID_DEVICE;

	close_device_fp(sci_device->handle); //execute the pointer function 


	if (cal_data != NULL)
		mj2adv_DeallocateCalDataMemory(cal_data); //free up cal data memory;
	free(device_data); //free device data allocated in open_device()
	free(sci_device); //free the device created in open_device()

	return EXIT_SUCCESS;
}

#ifdef __cplusplus
}
#endif