/*
 * MockingJay2 functions
 *	
 * Copyright (c) 2024 SignalCore Inc.
 *	
 *	Rev 2.2.0
*/



#ifdef __cplusplus
extern "C"
{
#endif
#include <stdlib.h>
#include <malloc.h>
#include <stdint.h>
#include <string.h>
#include <math.h>

#include "sci_types.h"
#include "sci_errors.h"
#include "sci_device.h"
#include "sci_math.h"

#include "mj2_comm.h"
#include "mj2_defs.h"
#include "mj2_regs.h"
#include "mj2_functions.h"
#include "mj2_advanceregs.h"
#include "mj2_device.h"
#include "mj2_advance.h" 
#include <math.h>

/*
	Begin product Export wrapper functions
	=================================================================================================================
*/
/* Implementation of Register access functions */
SCISTATUS mj2_RegWrite(HANDLE dev_handle, 
							uint8_t reg_byte,
							uint64_t instruct_word)
 {
	PSCIDEVICE sciDevice = (PSCIDEVICE)dev_handle;

	uint8_t bufferOut[TRANSFER_BYTES];
	uint8_t bufferIn[TRANSFER_BYTES];
	int i;
	int dataBytesOut = 0;
	SCISTATUS status = 0;

	switch(reg_byte)
	{
		case INITIALIZE: //Initialize Device
		case SYSTEM_ACTIVE: //Set System Active
		case SYNTH_MODE: //Set Synth Mode
		case RF_AMP: //-Enable RF amplifier
		case SIGNAL_PATH: //sets up the signal path of the converter
		case STORE_DEFAULT_STATE: //Stores the current device configuration
		case DEVICE_STANDBY: //Puts the entire or sections of the device in standby mode
		case REFERENCE_CLOCK: //Configure reference clock parameters
		case LO_SOURCE: //Set the drive path of the  signal, internal mixer or external mixer
		case SYNTH_SELF_CAL:
			dataBytesOut = 1;
			break;
		case ATTENUATOR: //Set the value of the attenuators
		case REFERENCE_DAC_WORD: //Adjust the reference clock via a DAC (calling STORE_DEFAULT_STATE reg make this value default)
		case USER_EEPROM_WRITE:
			dataBytesOut = 3;
			break;
		case AUTO_CALC_GAIN:
			dataBytesOut = 5;
			break;
		case RF_FREQUENCY: //Set RF frequency
		case IF_FREQUENCY: //Set IF frequency
		case LO_FREQUENCY:
			dataBytesOut = 7;
			break;

			//  Development Registers
		case DDS_WORD_32H:						//direct programming of DSS
		case DDS_WORD_32L:						//direct programming of DDS
		case LO_CRS_PLL_REG:					//direct programming of LO_CRS pll
		case LO_FINE_PLL_REG:					//direct programming of LO_Fine pll
		case SUM_CAL_FREQ:
			dataBytesOut = 5;
			break;
		case DDS_INSTRUCT_REG:					//direct programming of DDS
		case LO_SUM_PLL_REG:					//direct programming of LO_SUM_PLL
		case REF_100M_PLL_REG:
		case REF_10M_PLL_REG:
		case CAL_EEPROM_WRITE_REG:
			dataBytesOut = 3;
			break;
		case SET_LO_DAC_REG:
			dataBytesOut = 2;
			break;
		case EEPROMWP_REG:
			dataBytesOut = 1;
			break;
		default:
			dataBytesOut = -1;
			break;
	}

	if (!(dataBytesOut < 0))
	{
		bufferOut[0] = reg_byte;
		i = 0;
		while (i < dataBytesOut)
		{
			bufferOut[i + 1] = (uint8_t)((instruct_word >> (8 * (dataBytesOut - i - 1))) & 0xFF);
			i++;
		}

		if (sciDevice->device_type == USB_INT)
		{
			status = WriteDevice(dev_handle, bufferOut, sizeof(bufferOut));
			if (status != EXIT_SUCCESS) return status;
			status = ReadDevice(dev_handle, bufferIn, sizeof(bufferIn)); //read in always in 8 bytes
			if (status != EXIT_SUCCESS) return status;
		}
		else
		{
			status = WriteDevice(dev_handle, bufferOut, dataBytesOut + 1);
			if (status != SCI_SUCCESS) return status;

			status = ReadDevice(dev_handle, bufferIn, 1); //read in 1 acknowledge byte
			if (status != EXIT_SUCCESS) return status;
			if (bufferIn[0] != 1)
				return (SCISTATUS)SCI_ERROR_TIMED_OUT;
		}
	}
	else 
		return SCI_ERROR_INVALID_ARGUMENT;

	
	return EXIT_SUCCESS;
 }
	
SCISTATUS mj2_RegRead(HANDLE dev_handle, 
							uint8_t reg_byte,
							uint64_t instruct_word,
							uint64_t *received_word)
{
	PSCIDEVICE sciDevice = (PSCIDEVICE)dev_handle;
	SCISTATUS status;
	uint8_t bufferOut[TRANSFER_BYTES];
	uint8_t bufferIn[TRANSFER_BYTES];

	int transferred = 0;

	int i;
	int dataBytesOut = 0;
	int dataBytesIn = 8;
	uint64_t tmpWord = 0;

	switch (reg_byte)
	{
	case GET_DEVICE_PARAM: //Fetch device status
	case GET_TEMPERATURE: //Fetch Temperature
	case GET_DEVICE_STATUS: //Fetch Device status
	case GET_DEVICE_INFO: //Fetch Device Info: SN,
	case GET_SUM_ADC_REG:
		dataBytesOut = 1;
		break;
	case CAL_EEPROM_READ: // calibration EEPROM data
	case USER_EEPROM_READ: //user EEPROM data
		dataBytesOut = 3;
		break;
	default:
		dataBytesOut = -1;
		break;
	}

	if (!(dataBytesOut < 0))
	{
		bufferOut[0] = reg_byte;
		i = 0;
		while (i < dataBytesOut)
		{
			bufferOut[i + 1] = (uint8_t)((instruct_word >> (8 * (dataBytesOut - i - 1))) & 0xFF);
			i++;
		}

		if (sciDevice->device_type == USB_INT)
		{
			status = WriteDevice(dev_handle, bufferOut, sizeof(bufferOut));
			if (status != EXIT_SUCCESS) return status;
			status = ReadDevice(dev_handle, bufferIn, sizeof(bufferIn)); //read in always in 8 bytes
			if (status != EXIT_SUCCESS) return status;
		}
		else
		{
			status = WriteDevice(dev_handle, bufferOut, dataBytesOut + 1);
			if (status != SCI_SUCCESS) return status;

			status = ReadDevice(dev_handle, bufferIn, dataBytesIn); //read 8 bytes of return data
			if (status != EXIT_SUCCESS) return status;
		}
	}
	else
	{
		*received_word = 0xFF;
		return SCI_ERROR_INVALID_ARGUMENT;
	}

	i = 0;
	while (i < dataBytesIn)
	{
		tmpWord = tmpWord | (((uint64_t)bufferIn[i]) << (dataBytesIn - i -1) *8); //MSB
		i++;
	}
	*received_word = tmpWord;

	return EXIT_SUCCESS;
}
 

 /* Implementation of configuration (or setting ) functions */
SCISTATUS mj2_InitDevice(HANDLE dev_handle, uint8_t mode)
{
	SCISTATUS status;
	status = mj2_RegWrite(dev_handle,INITIALIZE,(mode & 0x01));
	if(status != EXIT_SUCCESS) return status;	

	if(mode == 1)
		status = mj2adv_FetchRfParameters(dev_handle);

	return status;
}

SCISTATUS mj2_SetFrequency(HANDLE dev_handle, double freq)
{
	if (dev_handle == NULL)
		return SCI_ERROR_INVALID_DEVICE_HANDLE;;
	SCISTATUS status;
	uint64_t freq_u64;
	attenuator_t atten;
	
	device_data_t *device_data = (device_data_t*)((PSCIDEVICE)dev_handle)->device_data;
	gain_params_t gain_params = device_data->gain_params;

	if (freq > 28.0e9) freq = 28.0e9;
	device_data->rf_params.rf_frequency = freq;
	if (device_data->rf_params.path_params.invert_spectrum)
	{
		device_data->rf_params.lo_frequency = device_data->rf_params.rf_frequency \
			+ device_data->rf_params.if_frequency;
	}
	else
	{
		device_data->rf_params.lo_frequency = device_data->rf_params.rf_frequency \
			- device_data->rf_params.if_frequency;
	}

	freq_u64 = (uint64_t)round(freq*1000.0);
	status = mj2_RegWrite(dev_handle, RF_FREQUENCY, freq_u64);
	if (status != EXIT_SUCCESS) return status;

	if ((device_data->auto_conv_enable) && (fabs(freq - device_data->last_rf) > 200e6))
	{
		status = mj2adv_AutoCalcAttenValues(dev_handle, &atten);
		if (status != SCI_SUCCESS)
			return status;

		if ((gain_params.auto_amp_ctrl) && (device_data->last_amp_state != device_data->rf_params.path_params.rf_amp_enable))
		{
			status = mj2_SetPreamp(dev_handle, device_data->rf_params.path_params.rf_amp_enable);
			if (status != SCI_SUCCESS)
				return status;
			device_data->last_amp_state = device_data->rf_params.path_params.rf_amp_enable;
		}

		status = mj2_SetAttenuator(dev_handle, IF_ATTEN, atten.if_atten_value);
		if (status != SCI_SUCCESS)
			return status;

		status = mj2_SetAttenuator(dev_handle, RF_ATTEN, atten.rf_atten_value);
		if (status != SCI_SUCCESS)
			return status;

		device_data->last_rf = freq;
	}

	return EXIT_SUCCESS;
}

SCISTATUS mj2_SetIfFrequency(HANDLE dev_handle, double freq)
{
	if (dev_handle == NULL)
		return SCI_ERROR_INVALID_DEVICE_HANDLE;
	SCISTATUS status;
	uint64_t freq_u64;
	attenuator_t atten;
	device_data_t* device_data = (device_data_t*)((PSCIDEVICE)dev_handle)->device_data;
	gain_params_t gain_params = device_data->gain_params;

	if (freq > 6.0e9) freq = 6.0e9;

	device_data->rf_params.if_frequency = freq;

	if (device_data->rf_params.path_params.invert_spectrum)
	{
		device_data->rf_params.lo_frequency = device_data->rf_params.rf_frequency \
			+ device_data->rf_params.if_frequency;
	}
	else
	{
		device_data->rf_params.lo_frequency = device_data->rf_params.rf_frequency \
			- device_data->rf_params.if_frequency;
	}

	freq_u64 = (uint64_t)round(freq * 1000.0);
	status = mj2_RegWrite(dev_handle, IF_FREQUENCY, freq_u64);
	if (status != EXIT_SUCCESS) return status;

	if ((device_data->auto_conv_enable) && (fabs(freq - device_data->last_if) > 100e6))
	{
		status = mj2adv_AutoCalcAttenValues(dev_handle, &atten);
		if (status != SCI_SUCCESS)
			return status;

		if ((gain_params.auto_amp_ctrl) && (device_data->last_amp_state != device_data->rf_params.path_params.rf_amp_enable))
		{
			status = mj2_SetPreamp(dev_handle, device_data->rf_params.path_params.rf_amp_enable);
			if (status != SCI_SUCCESS)
				return status;
			device_data->last_amp_state = device_data->rf_params.path_params.rf_amp_enable;
		}

		status = mj2_SetAttenuator(dev_handle, IF_ATTEN, atten.if_atten_value);
		if (status != SCI_SUCCESS)
			return status;

		status = mj2_SetAttenuator(dev_handle, RF_ATTEN, atten.rf_atten_value);
		if (status != SCI_SUCCESS)
			return status;
		device_data->last_if = freq;
	}

	return EXIT_SUCCESS;
}

SCISTATUS mj2_SetLoFrequency(HANDLE dev_handle, double freq)
{
	if (dev_handle == NULL)
		return SCI_ERROR_INVALID_DEVICE_HANDLE;;
	SCISTATUS status;
	uint64_t freq_u64;
	attenuator_t atten;
	device_data_t* device_data = (device_data_t*)((PSCIDEVICE)dev_handle)->device_data;
	gain_params_t gain_params = device_data->gain_params;

	if (freq > 28.0e9) freq = 28.0e9;
	device_data->rf_params.lo_frequency = freq;
	if (device_data->rf_params.path_params.invert_spectrum)
	{
		device_data->rf_params.rf_frequency = device_data->rf_params.lo_frequency \
			- device_data->rf_params.if_frequency;
	}
	else
	{
		device_data->rf_params.rf_frequency = device_data->rf_params.lo_frequency \
			+ device_data->rf_params.if_frequency;
	}

	freq_u64 = (uint64_t)round(freq * 1000.0);
	status = mj2_RegWrite(dev_handle, LO_FREQUENCY, freq_u64);
	if (status != EXIT_SUCCESS) return status;

	if ((device_data->auto_conv_enable) && (fabs(device_data->rf_params.rf_frequency - device_data->last_if) > 100e6))
	{
		status = mj2adv_AutoCalcAttenValues(dev_handle, &atten);
		if (status != SCI_SUCCESS)
			return status;

		if ((gain_params.auto_amp_ctrl) && (device_data->last_amp_state != device_data->rf_params.path_params.rf_amp_enable))
		{
			status = mj2_SetPreamp(dev_handle, device_data->rf_params.path_params.rf_amp_enable);
			if (status != SCI_SUCCESS)
				return status;
			device_data->last_amp_state = device_data->rf_params.path_params.rf_amp_enable;
		}

		status = mj2_SetAttenuator(dev_handle, IF_ATTEN, atten.if_atten_value);
		if (status != SCI_SUCCESS)
			return status;

		status = mj2_SetAttenuator(dev_handle, RF_ATTEN, atten.rf_atten_value);
		if (status != SCI_SUCCESS)
			return status;
		device_data->last_rf = device_data->rf_params.rf_frequency;
	}

	return EXIT_SUCCESS;
}

SCISTATUS mj2_SetSynthMode(HANDLE dev_handle, uint8_t pllLoopGain)
{
	int status;
	status = mj2_RegWrite(dev_handle, SYNTH_MODE, (pllLoopGain & 0x03));
	if (status != SCI_SUCCESS) return status;
	return SCI_SUCCESS;
}

SCISTATUS mj2_SetLoSource(HANDLE dev_handle, uint8_t external_lo,
	uint8_t lo_mode, uint8_t lo_doubler, uint8_t ext_lo_path)
{
	int status;
	uint64_t instructWord;
	instructWord = (uint64_t)(((ext_lo_path & 0x01) << 3) | ((lo_doubler & 0x01) << 2)
		| ((lo_mode & 0x01) << 1) | (external_lo & 0x01));
	status = mj2_RegWrite(dev_handle, LO_SOURCE, instructWord);
	if (status != SCI_SUCCESS) return status;
	return SCI_SUCCESS;
}

SCISTATUS mj2_SetAttenuator(HANDLE dev_handle, uint8_t attenuator, float atten_value)
{
	if (dev_handle == NULL)
		return SCI_ERROR_INVALID_DEVICE_HANDLE;
	uint64_t tmp_atten_value;
	PSCIDEVICE sci_device = (PSCIDEVICE)dev_handle;
	device_data_t* device_data = (device_data_t*)sci_device->device_data;

	if (attenuator == IF_ATTEN) //IF attenator
	{
		atten_value = round(atten_value * 4) * 0.25;
	}
	else
	{
		atten_value = round(atten_value * 2) * 0.5;
	}
	
	if (atten_value > 30.0)
		atten_value = 30.0;

	tmp_atten_value = (uint64_t)floor(atten_value * 4.0) | (uint64_t)(attenuator) << 8;

	if (attenuator == IF_ATTEN) //IF attenator
	{		
		device_data->rf_params.atten.if_atten_value = atten_value;
	}
	else //rf atten
	{
		device_data->rf_params.atten.rf_atten_value = atten_value;
	}

	return mj2_RegWrite(dev_handle, ATTENUATOR, tmp_atten_value);
}

SCISTATUS mj2_SetSignalPath(HANDLE dev_handle,	const signal_path_params_t* signal_path_params)
{
	if (dev_handle == NULL)
		return SCI_ERROR_INVALID_DEVICE_HANDLE;

	PSCIDEVICE sci_device = (PSCIDEVICE)dev_handle;
	device_data_t* device_data = (device_data_t*)sci_device->device_data;
	uint64_t tmpU64;
	attenuator_t atten;

	tmpU64 = ((uint64_t)signal_path_params->bypass_conversion & 0x01) |
		((uint64_t)(signal_path_params->rf_amp_enable & 0x01) << 1) |
		((uint64_t)(signal_path_params->if_out_enable & 0x01) << 2) |
		((uint64_t)(signal_path_params->invert_spectrum & 0x01) << 3);

	device_data->rf_params.path_params.bypass_conversion = signal_path_params->bypass_conversion;
	device_data->rf_params.path_params.rf_amp_enable = signal_path_params->rf_amp_enable;
	device_data->rf_params.path_params.if_out_enable = signal_path_params->if_out_enable;
	device_data->rf_params.path_params.invert_spectrum = signal_path_params->invert_spectrum;

	return mj2_RegWrite(dev_handle, SIGNAL_PATH, tmpU64);

	if (device_data->auto_conv_enable)
	{
		mj2adv_AutoCalcAttenValues(dev_handle, &atten);
		mj2_SetAttenuator(dev_handle, RF_ATTEN, atten.rf_atten_value);
		mj2_SetAttenuator(dev_handle, IF_ATTEN, atten.if_atten_value);
	}
}

SCISTATUS mj2_SetPreamp(HANDLE dev_handle, uint8_t enable)
{
	if (dev_handle == NULL)
		return SCI_ERROR_INVALID_DEVICE_HANDLE;
	SCISTATUS status;
	PSCIDEVICE sci_device = (PSCIDEVICE)dev_handle;
	attenuator_t atten;
	device_data_t* device_data = (device_data_t*)((PSCIDEVICE)dev_handle)->device_data;
	gain_params_t gain_params = device_data->gain_params;

	status = mj2_RegWrite(dev_handle, RF_AMP, (uint64_t)enable);
	if (status != EXIT_SUCCESS) return status;

	device_data->last_amp_state = enable;
	device_data->rf_params.path_params.rf_amp_enable = enable;
	device_data->gain_params.auto_amp_ctrl = 0;

	if ((device_data->auto_conv_enable) && (fabs(device_data->rf_params.rf_frequency - device_data->last_if) > 100e6))
	{
		status = mj2adv_AutoCalcAttenValues(dev_handle, &atten);
		if (status != SCI_SUCCESS)
			return status;

		status = mj2_SetAttenuator(dev_handle, IF_ATTEN, atten.if_atten_value);
		if (status != SCI_SUCCESS)
			return status;

		status = mj2_SetAttenuator(dev_handle, RF_ATTEN, atten.rf_atten_value);
		if (status != SCI_SUCCESS)
			return status;
		device_data->last_rf = device_data->rf_params.rf_frequency;
	}
	return EXIT_SUCCESS;
}

SCISTATUS mj2_SetStandby(HANDLE dev_handle, uint8_t standby_status)
{
	if (dev_handle == NULL)
		return SCI_ERROR_INVALID_DEVICE_HANDLE;
	SCISTATUS status;

	status = mj2_RegWrite(dev_handle, DEVICE_STANDBY, standby_status);
	if (status != EXIT_SUCCESS) return status;
	return EXIT_SUCCESS;
}

SCISTATUS mj2_SetReferenceClock(HANDLE dev_handle, uint8_t pxi10_clk_enable, uint8_t lock_ext_enable)
{
	uint64_t instructWord;

	int status;

	instructWord = (uint64_t)(((pxi10_clk_enable & 0x01) << 1) | (lock_ext_enable & 0x01));

	status = mj2_RegWrite(dev_handle, REFERENCE_CLOCK, instructWord);
	if (status != SCI_SUCCESS) return status;

	return SCI_SUCCESS;
}

SCISTATUS mj2_SetReferenceDac(HANDLE dev_handle, uint32_t dac_value)
{
	return mj2_RegWrite(dev_handle, REFERENCE_DAC_WORD, (uint64_t)(dac_value & 0x3FFF));
}

SCISTATUS mj2_StoreDefaultState(HANDLE dev_handle)
{
	if (dev_handle == NULL)
		return SCI_ERROR_INVALID_DEVICE_HANDLE;
	SCISTATUS status;
	PSCIDEVICE sci_device = (PSCIDEVICE)dev_handle;
	device_data_t* device_data = (device_data_t*)sci_device->device_data;

	status = mj2adv_SaveAutoConvParams(dev_handle, &device_data->gain_params, device_data->auto_conv_enable);
	if (status != EXIT_SUCCESS) return status;
	status = mj2_RegWrite(dev_handle, STORE_DEFAULT_STATE, 0x00);
	if (status != EXIT_SUCCESS) return status;
	return EXIT_SUCCESS;
}

SCISTATUS mj2_SetAutoConversion(HANDLE dev_handle, const gain_params_t *gain_params, uint8_t auto_conv_enable)
{
	uint64_t tmpU64;
	uint8_t tmpU8;
	SCISTATUS status;
	attenuator_t atten;

	PSCIDEVICE sci_device = (PSCIDEVICE)dev_handle; //cast the void* dev_handle memory back to a SCIDEVICE* struct
	device_data_t* device_data = (device_data_t*)sci_device->device_data;

	tmpU64 = ((uint64_t)auto_conv_enable & 0x01)
		| (((uint64_t)gain_params->auto_amp_ctrl & 0x01) << 1)
		| (((uint64_t)gain_params->linearity_mode & 0x07) << 2);


	tmpU8 = (uint8_t)floor(fabsf(gain_params->rf_level));
	if (gain_params->rf_level < 0)
		tmpU8 |= 0x80;
	else
		tmpU8 |= 0x00;
	tmpU64 |= ((uint64_t)tmpU8 << 24);

	tmpU8 = (uint8_t)floor(fabsf(gain_params->mixer_level));
	if (gain_params->mixer_level < 0)
		tmpU8 |= 0x80;
	else
		tmpU8 |= 0x00;
	tmpU64 |= ((uint64_t)tmpU8 << 16);

	tmpU8 = (uint8_t)round(fabsf(gain_params->if_level));
	if (gain_params->if_level < 0)
		tmpU8 |= 0x80;
	else
		tmpU8 |= 0x00;
	tmpU64 |= ((uint64_t)tmpU8 << 8);

	device_data->auto_conv_enable = auto_conv_enable;
	device_data->gain_params.rf_level = gain_params->rf_level;
	device_data->gain_params.mixer_level = gain_params->mixer_level;
	device_data->gain_params.if_level = gain_params->if_level;
	device_data->gain_params.linearity_mode = gain_params->linearity_mode;
	device_data->gain_params.auto_amp_ctrl = gain_params->auto_amp_ctrl;
	device_data->gain_params_set = 1;

	status = mj2_RegWrite(dev_handle, AUTO_CALC_GAIN, tmpU64);
	if (status != SCI_SUCCESS)
		return status;

	if (device_data->auto_conv_enable)
	{
		status = mj2adv_AutoCalcAttenValues(dev_handle, &atten);
		if (status != SCI_SUCCESS)
			return status;

		if (gain_params->auto_amp_ctrl)
		{
			status = mj2_SetPreamp(dev_handle, device_data->rf_params.path_params.rf_amp_enable);
			if (status != SCI_SUCCESS)
				return status;
			device_data->last_amp_state = device_data->rf_params.path_params.rf_amp_enable;
		}

		status = mj2_SetAttenuator(dev_handle, IF_ATTEN, atten.if_atten_value);
		if (status != SCI_SUCCESS)
			return status;

		status = mj2_SetAttenuator(dev_handle, RF_ATTEN, atten.rf_atten_value);
		if (status != SCI_SUCCESS)
			return status;
	}

	return SCI_SUCCESS;
}

SCISTATUS mj2_SetAsDefault(HANDLE dev_handle)
{
	return mj2_RegWrite(dev_handle, STORE_DEFAULT_STATE, 0);
}

SCISTATUS mj2_WriteUserEeprom(HANDLE dev_handle, uint32_t mem_add,	uint8_t byte_data)
{
	uint64_t instructWord;

	instructWord = ((uint64_t)(mem_add & 0xFFFF) << 8) | (uint64_t)byte_data | 1L << 24;
	return mj2_RegWrite(dev_handle, USER_EEPROM_WRITE, instructWord);
}

SCISTATUS mj2_SetSynthSelfCal(HANDLE dev_handle)
{
	return mj2_RegWrite(dev_handle, SYNTH_SELF_CAL, 0);
}

SCISTATUS mj2_SetGainCalcParams(HANDLE dev_handle, gain_params_t *gain_params)
{
	uint64_t tmpU64;
	uint8_t tmpU8;

	PSCIDEVICE sci_device = (PSCIDEVICE)dev_handle; //cast the void* dev_handle memory back to a SCIDEVICE* struct
	device_data_t* device_data = (device_data_t*)sci_device->device_data;

	tmpU64 = ((uint64_t)device_data->auto_conv_enable & 0x01)
		| (((uint64_t)gain_params->auto_amp_ctrl & 0x01) << 1)
		| (((uint64_t)gain_params->linearity_mode & 0x07) << 2);


	tmpU8 = (uint8_t)floor(fabsf(gain_params->rf_level));
	if (gain_params->rf_level < 0)
		tmpU8 |= 0x80;
	else
		tmpU8 |= 0x00;
	tmpU64 |= ((uint64_t)tmpU8 << 24);

	tmpU8 = (uint8_t)floor(fabsf(gain_params->mixer_level));
	if (gain_params->mixer_level < 0)
		tmpU8 |= 0x80;
	else
		tmpU8 |= 0x00;
	tmpU64 |= ((uint64_t)tmpU8 << 16);

	tmpU8 = (uint8_t)round(fabsf(gain_params->if_level));
	if (gain_params->if_level < 0)
		tmpU8 |= 0x80;
	else
		tmpU8 |= 0x00;
	tmpU64 |= ((uint64_t)tmpU8 << 8);

	device_data->gain_params.rf_level = gain_params->rf_level;
	device_data->gain_params.mixer_level = gain_params->mixer_level;
	device_data->gain_params.if_level = gain_params->if_level;
	device_data->gain_params.linearity_mode = gain_params->linearity_mode;
	device_data->gain_params.auto_amp_ctrl = gain_params->auto_amp_ctrl;
	device_data->gain_params_set = 1;

	return mj2_RegWrite(dev_handle, AUTO_CALC_GAIN, tmpU64);
}

/* begin implemenation of export Read functions */

SCISTATUS mj2_GetRfParameters(HANDLE dev_handle, rf_params_t *rf_params)
{
	if (dev_handle == NULL)
		return SCI_ERROR_INVALID_DEVICE_HANDLE;

	SCISTATUS status;
	uint64_t received_word = 0;  

	if (rf_params == NULL) return SCI_ERROR_MEM_UNALLOCATE; // if type is not allocated

	PSCIDEVICE sci_device = (PSCIDEVICE)dev_handle; //cast the void* dev_handle memory back to a SCIDEVICE* struct
	device_data_t *device_data = (device_data_t*)sci_device->device_data; //cast the void* device_data to a device_data_t* struct


	if (!device_data->rf_params_read)
	{
		status = mj2adv_FetchRfParameters(dev_handle);
		if (status != SCI_SUCCESS)
			return status;
	}

	*rf_params = device_data->rf_params;

	return EXIT_SUCCESS;
}

SCISTATUS mj2_GetGainCalcParams(HANDLE dev_handle, gain_params_t *gain_params)
{
	if (dev_handle == NULL)
		return SCI_ERROR_INVALID_DEVICE_HANDLE;

	SCISTATUS status;
	uint64_t received_word = 0;

	if (gain_params == NULL) return SCI_ERROR_MEM_UNALLOCATE; // if type is not allocated

	PSCIDEVICE sci_device = (PSCIDEVICE)dev_handle; //cast the void* dev_handle memory back to a SCIDEVICE* struct
	device_data_t* device_data = (device_data_t*)sci_device->device_data; //cast the void* device_data to a device_data_t* struct


	if (!device_data->gain_params_set)
	{
		status = mj2adv_FetchGainCalcParams(dev_handle);
		if (status != SCI_SUCCESS)
			return status;
		//printf("rf_params read\n");
	}

	*gain_params = device_data->gain_params;

	return EXIT_SUCCESS;
}

SCISTATUS mj2_GetTemperature(HANDLE dev_handle, float *temp)
{
	if (dev_handle == NULL)
		return SCI_ERROR_INVALID_DEVICE_HANDLE;
	SCISTATUS status;
	uint64_t received_word = 0;
	uint32_t tmp_u32;
	PSCIDEVICE sci_device = (PSCIDEVICE)dev_handle; //cast the void* dev_handle memory back to a SCIDEVICE* struct
	device_data_t *device_data = (device_data_t*)sci_device->device_data; //cast the void* device_data to a device_data_t* struct

	status = mj2_RegRead(dev_handle, GET_TEMPERATURE, 0, &received_word);
	if (status != EXIT_SUCCESS) return status;
	tmp_u32 = (uint32_t)received_word;
	*temp = *(float *)&tmp_u32;
	device_data->temperature = *temp; //copy temp into device_data memory

	return EXIT_SUCCESS;
}

SCISTATUS mj2_GetDeviceStatus(HANDLE dev_handle, device_status_t *device_status)
{
	if (dev_handle == NULL)
		return SCI_ERROR_INVALID_DEVICE_HANDLE;
	SCISTATUS status;
	uint64_t received_word = 0;

	if (device_status == NULL) return SCI_ERROR_MEM_UNALLOCATE; // if type is not allocated

	status = mj2_RegRead(dev_handle, GET_DEVICE_STATUS, 0, &received_word);
	if (status != EXIT_SUCCESS) return status;

	device_status->operate_status.pxi10_clk_enable = (uint8_t)((received_word & 0x0000800000) >> 23);
	device_status->operate_status.auto_amp_enable = (uint8_t)((received_word & 0x0000400000) >> 22);
	device_status->operate_status.auto_gain_enable = (uint8_t)((received_word & 0x0000200000) >> 21);
	device_status->operate_status.rf_amp_enable = (uint8_t)((received_word & 0x0000100000) >> 20);
	device_status->operate_status.invert_spect = (uint8_t)((received_word & 0x0000080000) >> 19);
	device_status->operate_status.if_out_enable = (uint8_t)((received_word & 0x0000040000) >> 18);
	device_status->operate_status.bypass_conv = (uint8_t)((received_word & 0x0000020000) >> 17);
	device_status->operate_status.device_standby = (uint8_t)((received_word & 0x0000010000) >> 16);
	device_status->operate_status.lo_doubler_enable = (uint8_t)((received_word & 0x0000008000) >> 15);
	device_status->operate_status.lo_mode = (uint8_t)((received_word & 0x0000004000) >> 14);
	device_status->operate_status.ext_lo_path = (uint8_t)((received_word & 0x0000002000) >> 13);
	device_status->operate_status.ext_lo_enable = (uint8_t)((received_word & 0x0000001000) >> 12);
	device_status->operate_status.lo_supply_enable = (uint8_t)((received_word & 0x0000000800) >> 11);
	device_status->operate_status.lock_ext_ref_enable = (uint8_t)((received_word & 0x0000000400) >> 10);
	device_status->operate_status.ext_ref_detected = (uint8_t)((received_word & 0x0000000200) >> 9);
	device_status->operate_status.device_accessed = (uint8_t)((received_word & 0x0000000100) >> 8);

	device_status->pll_status.loop_gain = (uint8_t)((received_word & 0x0000000060) >> 5);
	device_status->pll_status.tcxo_pll_ld = (uint8_t)((received_word & 0x0000000010) >> 4);
	device_status->pll_status.vcxo_pll_ld = (uint8_t)((received_word & 0x0000000008) >> 3);
	device_status->pll_status.fine_pll_ld = (uint8_t)((received_word & 0x0000000004) >> 2);
	device_status->pll_status.crs_pll_ld = (uint8_t)((received_word & 0x0000000002) >> 1);
	device_status->pll_status.sum_pll_ld = (uint8_t)(received_word & 0x0000000001);
	return EXIT_SUCCESS;
}

SCISTATUS mj2_GetDeviceInfo(HANDLE dev_handle, device_info_t *device_info)
{
	if (dev_handle == NULL)
		return SCI_ERROR_INVALID_DEVICE_HANDLE;
	SCISTATUS status;
	uint64_t received_word = 0;
	if (device_info == NULL) return SCI_ERROR_MEM_UNALLOCATE; // if type is not allocated

	PSCIDEVICE sci_device = (PSCIDEVICE)dev_handle; //cast the void* dev_handle memory back to a SCIDEVICE* struct
	device_data_t* device_data = (device_data_t*)sci_device->device_data; //cast the void* device_data to a device_data_t* struct


	if (!device_data->info_read)
	{
		status = mj2adv_FetchDeviceInfo(dev_handle);
		if (status != SCI_SUCCESS)
			return status;
	}

	*device_info = device_data->info;
	return EXIT_SUCCESS;
}

SCISTATUS mj2_GetCalData(HANDLE dev_handle, cal_data_t** cal_data)
{
	SCISTATUS status;
	PSCIDEVICE sci_device = (PSCIDEVICE)dev_handle; //cast the void* dev_handle memory back to a SCIDEVICE* struct
	device_data_t* device_data = (device_data_t*)sci_device->device_data; //cast the void* device_data to a device_data_t* 

	if (!device_data->cal_data_read)
	{
		status = mj2adv_FetchCalData(dev_handle);  // if cal data is not read yet fetch it
		if (status != SCI_SUCCESS)
			return status;
	}
	*cal_data = device_data->cal_data;

	return SCI_SUCCESS;
}

SCISTATUS mj2_CalcAttenValues(HANDLE dev_handle, attenuator_t* atten)
{
	return mj2adv_AutoCalcAttenValues(dev_handle, atten);
}

SCISTATUS mj2_CalcGain(HANDLE dev_handle, float *conversion_gain)
{
	SCISTATUS status;
	PSCIDEVICE sci_device = (PSCIDEVICE)dev_handle;
	cal_data_t* cal_data;
	device_data_t* device_data = (device_data_t*)sci_device->device_data;

	double ifFrequency = device_data->rf_params.if_frequency;
	double rfFrequency = device_data->rf_params.rf_frequency;
//	attenuator_t atten = device_data->atten;

	if (!device_data->cal_data_read)
	{
		status = mj2adv_FetchCalData(dev_handle);  // if cal data is not read yet, allocate and fetch
		if (status != EXIT_SUCCESS)
			return status;
	}

	cal_data = (cal_data_t*)device_data->cal_data; //use only data is already fetched;

	int rfIndex, ifIndex, rfAttenIndex, ifAttenIndex;
	double rfAbsGain, rfAmpGain, rfAtten, residueAtten;
	double ifGain, ifAtten;
	double deltaTemp, tempGain, tempcoeff1, tempcoeff2;

	ifAttenIndex = (int)round(device_data->rf_params.atten.if_atten_value);
	rfAttenIndex = (int)round(device_data->rf_params.atten.rf_atten_value) - 1;

	residueAtten = (device_data->rf_params.atten.if_atten_value - ifAttenIndex) + (device_data->rf_params.atten.rf_atten_value - (rfAttenIndex + 1));

	if (ifAttenIndex > (CALATTENSTEPS)) ifAttenIndex = (CALATTENSTEPS);

	if (rfAttenIndex > (CALATTENSTEPS - 1)) rfAttenIndex = (CALATTENSTEPS - 1);

	ifIndex = (int)floor(ifFrequency / 100e6);

	if (device_data->rf_params.path_params.bypass_conversion)
	{
		rfIndex = floor(rfFrequency / 100e6);
		*conversion_gain = cal_data->bp_abs_gain_cal[rfIndex];
	}
	else
	{
		rfIndex = floor((rfFrequency - 6e9) / 250e6);

		//step 1 determine the absolute rf gain
		if (device_data->rf_params.path_params.invert_spectrum)
			rfAbsGain = cal_data->rf_lsb_abs_gain_cal[rfIndex];
		else
			rfAbsGain = cal_data->rf_usb_abs_gain_cal[rfIndex];

		//step 2 determine the rf preamp gain if enabled
		if (device_data->rf_params.path_params.rf_amp_enable)
			rfAmpGain = cal_data->rf_amp_gain_cal[rfIndex];
		else
			rfAmpGain = 0;

		//step 3 determine the relative rf and if attenuation from cal values
		if (rfAttenIndex < 0)
			rfAtten = 0;
		else
			rfAtten = cal_data->rf_atten_cal[rfAttenIndex][rfIndex];
		if (ifAttenIndex < 0)
			ifAtten = 0;
		else
			ifAtten = cal_data->if_atten_cal[ifAttenIndex];

		//step 3 determine the ifGain relative to the 1 GHz point
		ifGain = cal_data->if_rel_gain_cal[ifIndex];

		//step 4 determine the temperature effects on gain
		deltaTemp = (double)(device_data->temperature - cal_data->cal_temp);

		if (rfFrequency > 20e9)
		{
			tempcoeff1 = cal_data->temp_coeff[4];
			tempcoeff2 = cal_data->temp_coeff[5];
		}
		else if (rfFrequency > 13e9)
		{
			tempcoeff1 = cal_data->temp_coeff[2];
			tempcoeff2 = cal_data->temp_coeff[3];
		}
		else
		{
			tempcoeff1 = cal_data->temp_coeff[0];
			tempcoeff2 = cal_data->temp_coeff[1];
		}

		tempGain = tempcoeff1 * deltaTemp + tempcoeff2 * pow(deltaTemp, 2);

		// Step 5 sum up  all gains and losses

		*conversion_gain = (float)(rfAbsGain + rfAmpGain + tempGain + ifGain - rfAtten - ifAtten - residueAtten);
	}

	return EXIT_SUCCESS;
}

SCISTATUS mj2_CalcMaxGain(HANDLE dev_handle, float* max_gain, float* pre_amp_gain)
{
	SCISTATUS status;
	PSCIDEVICE sci_device = (PSCIDEVICE)dev_handle;
	cal_data_t* cal_data;
	device_data_t* device_data = (device_data_t*)sci_device->device_data;

	double ifFrequency = device_data->rf_params.if_frequency;
	double rfFrequency = device_data->rf_params.rf_frequency;
	//	attenuator_t atten = device_data->atten;

	if (!device_data->cal_data_read)
	{
		status = mj2adv_FetchCalData(dev_handle);  // if cal data is not read yet, allocate and fetch
		if (status != EXIT_SUCCESS)
			return status;
	}

	cal_data = (cal_data_t*)device_data->cal_data; //use only data is already fetched;

	int rfIndex;

	rfIndex = floor((rfFrequency - 6e9) / 250e6);

	//step 1 determine the absolute rf gain
	if (device_data->rf_params.path_params.invert_spectrum)
		*max_gain = cal_data->rf_lsb_abs_gain_cal[rfIndex];
	else
		*max_gain = cal_data->rf_usb_abs_gain_cal[rfIndex];

	*pre_amp_gain = cal_data->rf_amp_gain_cal[rfIndex];


	return EXIT_SUCCESS;
}


 
#ifdef __cplusplus
}
#endif