/*
 *
  ******************************************************************************
 * @attention
 *
 * Copyright (c) 2020-2023 SignalCore.
 * All rights reserved.
 *
 * This software is licensed under terms that can be found in the LICENSE file
 * in the root directory of this software component.
 * If no LICENSE file comes with this software, it is provided AS-IS.
 *
 ******************************************************************************
 */

/*
* usbdevice.c : Defines the usb communication functions.
* 
* Rev 1.8
*/

#include <string.h>
#include <stdlib.h>
#include <stdint.h>
#include <libusb.h>

#include "sci_types.h"
#include "sci_device.h"
#include "sci_errors.h"
#include "sci_usb_device.h"



typedef struct _USBDEVICE
{
	libusb_device_handle *handle;
	libusb_context *usb_ctx;
} USBDEVICE, *PUSBDEVICE;

SCISTATUS UsbWrite(HANDLE deviceHandle, uint8_t *data, uint32_t len) 
{
	PUSBDEVICE usbDevice = (PUSBDEVICE)(deviceHandle);
	int transferred = 0;

	if (0 > libusb_bulk_transfer(usbDevice->handle, SCI_ENDPOINT_OUT_BULK, data, (int)len, &transferred, USB_TIMEOUT))
	{
		return SCI_ERROR_COMM_FAIL;
	}
	return SCI_SUCCESS;
}

SCISTATUS UsbRead(HANDLE deviceHandle, uint8_t *data, uint32_t len)
{
	PUSBDEVICE usbDevice = (PUSBDEVICE)deviceHandle;
	int transferred = 0;

	if (0 > libusb_bulk_transfer(usbDevice->handle, SCI_ENDPOINT_IN_BULK, data, (int)len, &transferred, USB_TIMEOUT))
	{
		return SCI_ERROR_COMM_FAIL;
	}

	return SCI_SUCCESS;
}

SCISTATUS UsbSearchDevices(char **serialNumberList, uint32_t *numberDevices)
{
	libusb_device_handle *devHandle = NULL;
	int deviceCount = 0;
	uint8_t string_usb[9];
	libusb_device *device = NULL;
	libusb_context *ctx = NULL;
	libusb_device **devList;
	struct libusb_device_descriptor desc;
	int r, i;
	int status = 0;
	ssize_t count;

	if (serialNumberList == NULL)
	{
		return SCI_ERROR_MEM_UNALLOCATE;
	}

	r = libusb_init(&ctx);
	if (r < 0) return SCI_ERROR_COMM_INIT;

	count = libusb_get_device_list(ctx, &devList);
	if (count < 0)
	{
		libusb_free_device_list(devList, 1);
		return SCI_ERROR_NO_DEVICE;
	}
	for (i = 0; i < count; i++)
	{
		device = devList[i];;
		r = libusb_get_device_descriptor(device, &desc);
		if (desc.idVendor == SCI_USB_VID && desc.idProduct == SCI_USB_PID)
		{
			status = libusb_open(device, &devHandle);
			if (status == 0)
			{
				libusb_get_string_descriptor_ascii(devHandle, desc.iSerialNumber, string_usb, sizeof(string_usb));
				strncpy(serialNumberList[deviceCount], (const char*)string_usb, SCI_SN_LENGTH);
				libusb_close(devHandle);
				deviceCount++;
			}
		}
	}
	libusb_free_device_list(devList, 1);
	libusb_exit(ctx);
	*numberDevices = deviceCount;
	return SCI_SUCCESS;
}

SCISTATUS UsbSearchDevicesLV(char *serialNumberList, uint32_t *numberDevices)
{
	libusb_device_handle *deviceHandle = NULL;
	int deviceCount = 0;
	uint8_t string_usb[9];
	char *list;
	libusb_device **devList;
	libusb_context *ctx = NULL;
	struct libusb_device_descriptor desc;
	int r, i;
	int status = 0;
	ssize_t count;

	list = (char*)malloc(160);
	if (list == NULL) return 0;
	list[0] = '\0';
	r = libusb_init(&ctx);
	if (r < 0) return SCI_ERROR_COMM_INIT;

	count = libusb_get_device_list(ctx, &devList);
	if (count < 0)
	{
		libusb_free_device_list(devList, 1);
		return SCI_ERROR_NO_DEVICE;
	}
	for (i = 0; i < count; i++)
	{
		status = libusb_get_device_descriptor(devList[i], &desc);
		if (desc.idVendor == SCI_USB_VID && desc.idProduct == SCI_USB_PID)
		{
			status = libusb_open(devList[i], &deviceHandle);
			if (status == 0)
			{
				libusb_get_string_descriptor_ascii(deviceHandle, desc.iSerialNumber, string_usb, sizeof(string_usb));
				strncat(list, (char *)string_usb, SCI_SN_LENGTH + 1);
				libusb_close(deviceHandle);
				deviceCount++;
			}
		}
	}
	strcpy(serialNumberList, list);

	libusb_free_device_list(devList, 1);
	libusb_exit(ctx);
	free(list);

	*numberDevices = deviceCount;
	return SCI_SUCCESS;
}

SCISTATUS UsbOpenDevice(char *devSerialNum, __attribute__((unused)) uint8_t baudrate, PHANDLE deviceHandle)
{
	UNREFERENCED_PARAMETER(baudrate);
	PUSBDEVICE usbDevice;
	uint8_t string_usb[9];
	//	uint8_t *string_usb;
	libusb_device *device = NULL;
	libusb_device **devList;
	struct libusb_device_descriptor desc;
	libusb_device_handle *tmpHandle;
	uint8_t ledData[8];

	int i, r;
	ssize_t count;
	int status;

	usbDevice = (PUSBDEVICE)malloc(sizeof(USBDEVICE));

	//	string_usb = (uint8_t*)malloc(sizeof(char)*9);

	r = 0;
	status = libusb_init(&(usbDevice->usb_ctx));
	count = libusb_get_device_list(usbDevice->usb_ctx, &devList);
	for (i = 0; i < count; i++)
	{
		status = libusb_get_device_descriptor(devList[i], &desc);
		if (desc.idVendor == SCI_USB_VID && desc.idProduct == SCI_USB_PID)
		{
			status = libusb_open(devList[i], &tmpHandle);
			if (status == 0)
			{
				libusb_get_string_descriptor_ascii(tmpHandle, desc.iSerialNumber, string_usb, sizeof(string_usb));
				if (strncmp((const char*)string_usb, devSerialNum, SCI_SN_LENGTH) == 0)
				{
					device = devList[i];
					r = 1;
				}
				libusb_close(tmpHandle);
			}
		}
	}

	if (r)
	{
		if (libusb_open(device, &usbDevice->handle) == 0)
			status = libusb_reset_device(usbDevice->handle);
		libusb_set_configuration(usbDevice->handle, 1);
		if (!libusb_claim_interface(usbDevice->handle, 0)) libusb_set_interface_alt_setting(usbDevice->handle, 0, 1);
		ledData[0] = 0x02;
		ledData[1] = 0x01;
		status = UsbWrite((HANDLE*)usbDevice, ledData, sizeof(ledData));
		status = UsbRead((HANDLE*)usbDevice, ledData, sizeof(ledData));
		libusb_free_device_list(devList, 1);
		*deviceHandle = (HANDLE*)usbDevice;
		return status;
	}
	else
	{
		libusb_exit(usbDevice->usb_ctx);
		free(usbDevice);
		*deviceHandle = NULL;
		return SCI_ERROR_NO_DEVICE;
	}

	
}

SCISTATUS UsbOpenDeviceLV(char *devSerialNum, PHANDLE deviceHandle)
{
	PUSBDEVICE usbDevice;
	uint8_t string_usb[9];
	libusb_device *device = NULL;
	libusb_device **devList;
	struct libusb_device_descriptor desc;
	libusb_device_handle *tmpHandle;
	uint8_t ledData[8];

	int i, r;
	ssize_t count;
	int status;

	usbDevice = (PUSBDEVICE)malloc(sizeof(USBDEVICE));

	r = 0;
	status = libusb_init(&(usbDevice->usb_ctx));
	count = libusb_get_device_list(usbDevice->usb_ctx, &devList);
	for (i = 0; i < count; i++)
	{
		status = libusb_get_device_descriptor(devList[i], &desc);
		if (desc.idVendor == SCI_USB_VID && desc.idProduct == SCI_USB_PID)
		{
			status = libusb_open(devList[i], &tmpHandle);
			if (status == 0)
			{
				libusb_get_string_descriptor_ascii(tmpHandle, desc.iSerialNumber, string_usb, sizeof(string_usb));
				if (strncmp(string_usb, devSerialNum, SCI_SN_LENGTH) == 0)
				{
					device = devList[i];
					r = 1;
				}
				libusb_close(tmpHandle);
			}
		}
	}

	if (r)
	{
		if (libusb_open(device, &usbDevice->handle) == 0)
			status = libusb_reset_device(usbDevice->handle);
		libusb_set_configuration(usbDevice->handle, 1);
		if (!libusb_claim_interface(usbDevice->handle, 0)) libusb_set_interface_alt_setting(usbDevice->handle, 0, 1);
		ledData[0] = 0x02;
		ledData[1] = 0x01;
		status = UsbWrite((PHANDLE)usbDevice, ledData, sizeof(ledData));
		status = UsbRead((PHANDLE)usbDevice, ledData, sizeof(ledData));
		libusb_free_device_list(devList, 1);
		*deviceHandle = (PHANDLE)usbDevice;
		return status;
	}
	else
	{
		libusb_exit(usbDevice->usb_ctx);
		free(usbDevice);
		deviceHandle = 0;
		return SCI_ERROR_NO_DEVICE;
	}
}

SCISTATUS UsbCloseDevice(HANDLE deviceHandle)
{

	libusb_device *dev = NULL;
	PUSBDEVICE usbDevice = (PUSBDEVICE)deviceHandle;
	SCISTATUS status;
	char ledData[8];

	ledData[0] = 0x02;
	ledData[1] = 0x00;

	dev = libusb_get_device(usbDevice->handle);
	if (dev == NULL) return SCI_ERROR_INVALID_DEVICE;  //invalid handle	

	status = UsbWrite(deviceHandle, ledData, sizeof(ledData));
	status = UsbRead(deviceHandle, ledData, sizeof(ledData));
	if (SCI_SUCCESS != status)
		{
			return status;
		}

	libusb_release_interface(usbDevice->handle, 0);
	libusb_close(usbDevice->handle);
	libusb_exit(usbDevice->usb_ctx);
	free(usbDevice);
	return SCI_SUCCESS;
}